
//**************************************************************************
//**
//** CL_WORLD.C
//**
//** Client side world simulation.
//**
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "dd_def.h"
#include "dd_net.h"
#include "cl_def.h"
#include "console.h"

// MACROS ------------------------------------------------------------------

#define MAX_MOVERS			128		// Definitely enough!

#define MVF_CEILING			0x1		// Move ceiling.
#define MVF_SET_FLOORPIC	0x2		// Set floor texture when move done.

typedef enum
{
	mvt_floor,
	mvt_ceiling
} movertype_t;

// TYPES -------------------------------------------------------------------

typedef struct
{
	thinker_t	thinker;

	sector_t	*sector;
	int			sectornum;
	movertype_t	type;
	fixed_t		*current;
	fixed_t		destination;
	fixed_t		speed;
	int			floorpic;
	int			sound_id;
	int			sound_pace;
} mover_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static mover_t *activemovers[MAX_MOVERS];

// CODE --------------------------------------------------------------------

void Cl_InitMovers()
{
	int		i;

	for(i=0; i<MAX_MOVERS; i++) activemovers[i] = NULL;
}

void Cl_RemoveActiveMover(mover_t *mover)
{
	int		i;

	for(i=0; i<MAX_MOVERS; i++)
		if(activemovers[i] == mover)
		{
			// Stop the sound?
			if(mover->sound_id)
			{
				gx.NetWorldEvent(DDWE_SECTOR_SOUND, mover->sectornum << 16, NULL);
			}
			P_RemoveThinker(&mover->thinker);
			activemovers[i] = NULL;
			break;
		}
}

void Cl_MoverThinker(mover_t *mover)
{
	fixed_t	*current = mover->current;
	boolean remove = false;

	if(!Cl_GameReady()) return; // Can we think yet?

	// Should we play a sound?
	if(mover->sound_id && mover->sound_pace && 
		!(gametic % mover->sound_pace))
	{
		gx.NetWorldEvent(DDWE_SECTOR_SOUND, 
			(mover->sectornum << 16) | mover->sound_id, NULL);
	}

	// How's the gap?
	if(abs(mover->destination - *current) > abs(mover->speed))
	{
		// Do the move.
		*current += mover->speed;			
	}
	else
	{
		// We have reached the destination.
		*current = mover->destination;
		// Is a texture change in order?
		if(mover->floorpic >= 0)
		{
			mover->sector->floorpic = mover->floorpic;
		}
		// This thinker can now be removed.
		remove = true;
	}
	// Moving floors often act as lifts, etc. The ride should be smooth.
	if(mover->type == mvt_floor)
	{
		mobj_t *mo;
		// Adjust the mobjs linked to this sector.
		for(mo=mover->sector->thinglist; mo; mo=mo->snext)
		{
			if(mo->z < mover->sector->floorheight)
			{
				mo->z = mover->sector->floorheight;
				//mo->momz = mover->speed;
			}
		}
	}
	// Can we remove this thinker?
	if(remove) Cl_RemoveActiveMover(mover);
}

void Cl_AddMover(int sectornum, movertype_t type, fixed_t dest, fixed_t speed,
				 int floorpic)
{
	sector_t	*sector;
	int			i;
	mover_t		*mov;

	if(sectornum >= numsectors) return;
	sector = SECTOR_PTR(sectornum);

	// Remove any existing movers for the same plane.
	for(i=0; i<MAX_MOVERS; i++)
		if(activemovers[i] 
			&& activemovers[i]->sector == sector 
			&& activemovers[i]->type == type)
		{
			Cl_RemoveActiveMover(activemovers[i]);
		}

	if(!speed) // Lightspeed?
	{
		// Change immediately.
		if(type == mvt_ceiling)
			sector->ceilingheight = dest;
		else
			sector->floorheight = dest;
		if(floorpic >= 0) sector->floorpic = floorpic;
		return;
	}

	// Add a new mover.
	for(i=0; i<MAX_MOVERS; i++)
		if(activemovers[i] == NULL)
		{
			// Allocate a new mover_t thinker.
			mov = activemovers[i] = Z_Malloc(sizeof(mover_t), PU_LEVEL, 0);
			memset(mov, 0, sizeof(mover_t));
			mov->thinker.function = Cl_MoverThinker;			
			mov->type = type;
			mov->floorpic = floorpic;
			mov->sectornum = sectornum;
			mov->sector = SECTOR_PTR(sectornum);
			mov->destination = dest;
			mov->speed = speed; 
						
			mov->current = type==mvt_floor? 
				&mov->sector->floorheight : &mov->sector->ceilingheight;
			// Set the right sign for speed.
			if(mov->destination < *mov->current) mov->speed = -mov->speed;
			
			P_AddThinker(&mov->thinker);
			break;
		}
}

void Cl_RemoveMovers()
{
	int		i;

	for(i=0; i<MAX_MOVERS; i++)
		if(activemovers[i])
		{
			P_RemoveThinker(&activemovers[i]->thinker);
			activemovers[i] = NULL;
		}
}

void Cl_HandleSectorUpdate()
{
	fixed_t dest, speed;
	packet_sectorupdate_t upd;
	sector_t *sec;

	while(!Msg_End())
	{
		Msg_Read(&upd, sizeof(upd));
		if(upd.sector >= numsectors) break; // Bugger.
		sec = SECTOR_PTR(upd.sector);
		if(upd.flags & DDSU_FLOOR_HEIGHT)
		{
			sec->floorheight = Msg_ReadShort() << FRACBITS; 
		}
		if(upd.flags & DDSU_FLOOR_MOVING)
		{
			dest = Msg_ReadShort() << FRACBITS;
			speed = Msg_ReadShort() << 8;
			Cl_AddMover(upd.sector, mvt_floor, dest, speed, -1);
		}
		if(upd.flags & DDSU_FLOORPIC)
		{
			sec->floorpic = Msg_ReadShort();
		}	
		if(upd.flags & DDSU_CEILING_HEIGHT)
		{
			sec->ceilingheight = Msg_ReadShort() << FRACBITS;
		}			
		if(upd.flags & DDSU_CEILING_MOVING)
		{
			dest = Msg_ReadShort() << FRACBITS;
			speed = Msg_ReadShort() << 8;
			Cl_AddMover(upd.sector, mvt_ceiling, dest, speed, -1);
		}
		if(upd.flags & DDSU_CEILINGPIC)
		{
			sec->ceilingpic = Msg_ReadShort();
		}	
		if(upd.flags & DDSU_LIGHT_LEVEL)
		{
			sec->lightlevel = Msg_ReadByte();
		}
	}
}

void Cl_HandleWallUpdate()
{
	packet_wallupdate_t upd;
	side_t *side;

	while(!Msg_End())
	{
		Msg_Read(&upd, sizeof(upd));
		if(upd.side >= numsides) break; // Bugger.
		side = SIDE_PTR(upd.side);
		if(upd.flags & DDWU_TOP)
		{
			side->toptexture = Msg_ReadShort();
		}
		if(upd.flags & DDWU_MID)
		{
			side->midtexture = Msg_ReadShort();
		}
		if(upd.flags & DDWU_BOTTOM)
		{
			side->bottomtexture = Msg_ReadShort();
		}
	}
}

mover_t *Cl_GetActiveMover(int sectornum, movertype_t type)
{
	int i;

	for(i=0; i<MAX_MOVERS; i++)
		if(activemovers[i] 
			&& activemovers[i]->sectornum == sectornum
			&& activemovers[i]->type == type)
		{
			return activemovers[i];
		}
	return NULL;
}

void Cl_HandlePlaneSound()
{
	int sectornum = Msg_ReadShort();
	movertype_t type;
	int pace = Msg_ReadByte();
	mover_t *mov;

	type = pace & 0x80? mvt_ceiling : mvt_floor;
	// Try to find the mover.
	mov = Cl_GetActiveMover(sectornum, type);
	if(!mov) return;
	mov->sound_id = Msg_ReadShort();
	mov->sound_pace = pace & 0x7f;
}