#ifndef __DOOMSDAY_NETWORK_H__
#define __DOOMSDAY_NETWORK_H__

#include "dd_share.h"
#include "i_net.h"
#include "net_msg.h"

#define SV_VERSION		2

#define NSP_BROADCAST	-1			// For Net_SendPacket.

// Flags for console text from the server.
// Change with server version?
#define SV_CONSOLE_FLAGS	(CBLF_WHITE|CBLF_LIGHT|CBLF_GREEN)

#define NETBUFFER_MAXDATA	9215

#define PING_TIMEOUT		1000	// Ping timeout (ms).
#define MAX_PINGS			10

#define MONITORTICS			7

// The number of mobjs that can be stored in the input/visible buffer.
// The server won't send more mobjs than this.
#define	MAX_CLMOBJS		80	

// Packet types. 
// pkt = sent by anybody
// psv = sent by server
// pcl = sent by client
enum
{
	// Responses.
	pcl_hello		= 0,
	pkt_ok,			
	pkt_cancel,
	pkt_player_info,
	pkt_chat,
	pkt_ticcmd,
	pkt_ping,
	psv_handshake,
	psv_server_close,
	psv_frame,
	psv_player_exit,
	psv_console_text,

	// World events.
	psv_sector_update	= DDPT_SECTOR_UPDATE,		// 16
	psv_wall_update		= DDPT_WALL_UPDATE,			// 17
	psv_plane_sound		= DDPT_PLANE_SOUND,			// 18
	
	// Game specific events.
	pkt_game_marker		= DDPT_FIRST_GAME_EVENT		// 64
};

#define RESENDCOUNT			10
#define HANDSHAKECOUNT		17
#define UPDATECOUNT			5

// Mobj Data Packet flags (upper byte of the spritenum word).
#define MDP_EXTRA_FLAGS		0x02	
#define MDP_POSITION_Z		0x04	// Z coord included (otherwise on floor).
#define MDP_MOVEMENT_XY		0x08
#define MDP_MOVEMENT_Z		0x10
#define MDP_DIRECTION		0x20
#define MDP_FRAME			0x40
#define MDP_ACCURATE		0x80	// Accurate position and direction.

#define MDP_MOVEMENT		0x18
#define MDP_ALL				0x7e

#define MDP_SPRITE_MASK		0x01ff
#define MDP_FLAG_MASK		0xfe00
#define MDP_FRAME_MASK		0x7f

// Packed mobj flags (in the Extra Flags word).
#define MDPE_SHADOW			0x0001
#define MDPE_ALTSHADOW		0x0002
#define MDPE_BRIGHTSHADOW	0x0003	// Shadow and altshadow combined.
#define MDPE_FLOORCLIP		0x0004	// The floorclip byte will follow.
#define MDPE_NOFITBOTTOM	0x0008
#define MDPE_VIEWALIGN		0x0010
#define MDPE_SOLID			0x0020	// Link to blockmap.
#define MDPE_EXTENDED		0x0040	// An extended flags byte follows.
#define MDPE_LIGHTOFFSET	0x0780	// Light offset, along Z axis.
#define MDPE_TRANSCLASS		0x1800	// Translation class (>>MDPE_CLASSSHIFT)
#define MDPE_TRANSLATION	0xe000	// Translation table (>>MDPE_TRANSSHIFT)

// Extra flags, extended byte.
#define MDPE2_LIGHTSCALE	0x03	// Light scale (0: full, 3: 1/4).

#define MDPE_LIGHTSCALESHIFT	0
#define MDPE_LIGHTOFFSETSHIFT	7
#define MDPE_CLASSSHIFT			11
#define MDPE_TRANSSHIFT			13

#define DDMF_PACK_MASK		0x3c03f1fe	// These flags are packed.

// Packed frame flags (in the Frame byte).
#define MDPF_FULLBRIGHT		0x80

// Camera Data Packet flags (first byte).
#define CDP_FIX_ANGLES		0x01	// Camera angles included.
#define CDP_FILTER			0x02	// Filter included.
#define CDP_EXTRA_LIGHT		0x04


//---------------------------------------------------------------------------
// Types
//---------------------------------------------------------------------------

typedef struct
{
	// High tics when ping was sent (0 if pinger not used).
	int		sent;			

	// A record of the pings (negative time: no response).
	float	times[MAX_PINGS];	

	// Total number of pings and the current one.
	int		total;
	int		current;
} 
pinger_t;

// Network information for a player. Corresponds the players array.
typedef struct
{
	// Ticcmd buffer. The server uses this when clients send it ticcmds.
	byte	*ticcmds;		

	// Number of tics in the buffer.
	int		numtics;	

	// Index of the first tic in the buffer.
	int		firsttic;

	// If >0, the server will send the next world frame to the client.
	// This is set when input is received from the client. 
	int		updateCount;

	// Clients use this to determine how long ago they received the
	// last update of this client.
	int		age;

	// Is this client connected? (Might not be in the game yet.) Only
	// used by the server.
	int		connected;

	// If true, the server will send the player a handshake. The client must
	// acknowledge it before this flag is turned off.
	int		handshake;

	// Server uses this to determine whether it's OK to send game packets
	// to the client.
	int		ready;

	// The name of the player.
	char	name[PLAYERNAMELEN];

	// Field of view. Used in determining visible mobjs (default: 90).
	float	fov;

	// jtNet node for this player.
	unsigned int nodeID;	// jtNet node ID.
	int		jtNetNode;		// The corresponding jtNet node number (-1=none).

	// Ping tracker for this client.
	pinger_t ping;
} 
client_t;

typedef struct
{
	byte	type;
	byte	data[NETBUFFER_MAXDATA];
} netdata_t;

typedef struct
{
	int		player;		// Recipient or sender.
	int		length;		// Number of bytes in the data buffer.
	int		headerLength; // 1 byte at the moment.

	byte	*cursor;	// Points into the data buffer.

	// The data buffer for sending and receiving packets.
	netdata_t msg;

} netbuffer_t;

// Packets.
typedef struct
{
	byte				version;
	unsigned short		playerMask;		// 16 players.
	byte				yourConsole;	// Which one are you?
} handshake_packet_t;

typedef struct
{
	byte	console;
	char	name[PLAYERNAMELEN];
} playerinfo_packet_t;


//---------------------------------------------------------------------------
// Variables
//---------------------------------------------------------------------------
extern boolean firstRunTic, firstNetUpdate;
extern netbuffer_t netbuffer;
extern int resend_start;			// set when server needs our tics
extern int resend_count;
extern int gametime, oldentertics;
extern int num_clmobjs;
extern char *serverName, *serverInfo, *playerName;
extern char *masterAddress;
extern int masterPort;
extern boolean masterAware;


//---------------------------------------------------------------------------
// Functions
//---------------------------------------------------------------------------
void Net_AllocArrays(void);
void D_SendPacket(int to_player, int type, void *data, int length);
boolean Net_GetPacket(void);
void Net_SendPacket(int to_player, int sp_flags);
void Net_InitGame (void);
void Net_StartGame (void);
void Net_StopGame (void);
void Net_SendPing(int player, int count);
void Net_PingResponse(void);
void Net_ShowPingSummary(int player);

#endif