
//**************************************************************************
//**
//** I_NET.C
//**
//** Uses jtNet2.
//**
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "dd_def.h"
#include "sv_def.h"
#include "i_win32.h"
#include "i_timer.h"
#include "console.h"
#include "cl_def.h"
#include "sv_def.h"
#include "jtNetEx.h" // Exported functions from jtNet2.dll.

// MACROS ------------------------------------------------------------------

#define LFID_NONE			-1
#define DCN_MISSING_NODE	-10

#define DEFAULT_MASTER_PORT	10123	

#define MASTER_QUEUE_LEN	16

// TYPES -------------------------------------------------------------------

// If a master action fails, the action queue is emptied.
typedef enum 
{
	mac_connect,			// Connect to master.
	mac_check_connection,	// Wait for the connection to form.
	mac_disconnect,			// Disconnect from master.
	mac_get_servers,		// Retrieve the list of servers from the master.
	mac_wait_servers,		// Wait for the server list to arrive.
	mac_list_servers		// Print the server list in the console.
} masteraction_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

netbuffer_t	netbuffer;
int			maxQueuePackets = 0; //5;

char		*serverName = "Doomsday";
char		*serverInfo = "Multiplayer game server";
char		*playerName = "Player";

char		*masterAddress = "";
int			masterPort = DEFAULT_MASTER_PORT;		
boolean		masterAware = false;		

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static boolean netAvailable = false;

static int	lastFrameId = LFID_NONE;
static int	localtime;
static char *protocolNames[5] = { NULL, "IPX", "TCP/IP", "Serial Link", "Modem" };

// The master action queue. Perhaps not the best way to do this,
// but works nonetheless.
static masteraction_t masterQueue[MASTER_QUEUE_LEN];
static int	mqHead, mqTail;

// CODE --------------------------------------------------------------------

// Add a master action command to the queue.
void I_MAPost(masteraction_t act)
{
	masterQueue[mqHead] = act;
	mqHead = (++mqHead) % MASTER_QUEUE_LEN;
	//if(mqHead == mqTail) I_Error("I_MAPost: master action queue overflow!\n");
}

boolean I_MAGet(masteraction_t *act)
{
	// Empty queue?
	if(mqHead == mqTail) return false;	
	*act = masterQueue[mqTail];
	return true;
}

void I_MARemove(void)
{
	if(mqHead != mqTail) mqTail = (++mqTail) % MASTER_QUEUE_LEN;
}

void I_MAClear(void)
{
	mqHead = mqTail = 0;
}

int I_NetUpdateNodes()
{
	unsigned int idlist[MAXPLAYERS];
	int		i, k, numnodes = jtNetGetPlayerIDs(idlist);

	if(server)
	{
		clients[0].nodeID = jtNetGetInteger(JTNET_MY_PLAYER_ID);
		// Find the correct jtNet nodes for each player.
		for(i=0; i<MAXPLAYERS; i++)
			if(clients[i].connected)
			{
				clients[i].jtNetNode = DCN_MISSING_NODE;
				for(k=0; k<MAXPLAYERS; k++)			
					if(idlist[k] == clients[i].nodeID)
					{
						clients[i].jtNetNode = k;
						break;
					}
			}
	}
	else
	{
		// Clients only know about the server.
		clients[0].jtNetNode = JTNET_SERVER_NODE;
	}
	return numnodes;
}

// Returns the name of the specified player.
char *I_NetGetPlayerName(int player)
{
	return clients[player].name;
}

// Returns true if the operation is successful.
int I_NetSetPlayerData(void *data, int size)
{
	return jtNetSetData(JTNET_PLAYER_DATA, data, size);
}

// Returns true if the operation is successful.
int I_NetGetPlayerData(int playerNum, void *data, int size)
{
	int		node=-1;

	if(playerNum >= 0 && playerNum < MAXPLAYERS)
		node = clients[playerNum].jtNetNode;
		
/*	// Find the node for this player.
	for(i=0; i<MAXPLAYERS; i++)
		if(dcNodes[i].consoleplayer == playerNum)
		{
			node = dcNodes[i].jtNetNode;
			break;
		}
	if(node == -1) return false;
*/
	return jtNetGetData(JTNET_PLAYER_DATA, node, data, size);
}

int I_NetSetServerData(void *data, int size)
{
	return jtNetSetData(JTNET_SERVER_DATA, data, size);
}

int I_NetGetServerData(void *data, int size)
{
	return jtNetGetData(JTNET_SERVER_DATA, 0, data, size);
}

// Initialize the driver. Returns true if successful.
int I_NetInit(int sp)
{
	int		i;

	i = jtNetInit(sp);
	if(i == JTNET_ERROR_ALREADY_INITIALIZED) 
	{
		ST_Message("I_NetInit: Network driver already initialized.\n");
		return true;
	}
	if(i != JTNET_ERROR_OK)
		//I_Error("I_NetInit: Initialization failed.\n");
		return false;

	netAvailable = true;

	// Set some callback functions.
	jtNetSetCallback(JTNET_PLAYER_CREATED_CALLBACK, Sv_PlayerArrives);
	jtNetSetCallback(JTNET_PLAYER_DESTROYED_CALLBACK, Sv_PlayerLeaves);

	localtime = 0;

	// The game dll specifies a game id for us.
	jtNetSetString(JTNET_APPLICATION_NAME, gx.GetString(DD_GAME_ID));

	// Clear the master action queue.
	mqHead = mqTail = 0;
	return true;
}

void I_NetShutdown()
{
	if(netgame)
	{
		CON_Execute(server? "net server close" : "net disconnect", true);
	}
	netAvailable = false;
	netgame = false;
	server = false;
	jtNetShutdown();
}

// Send the data in the doomcom databuffer.
// destination is a player number.
void I_SendPacket(int flags)
{
	int msgId;
	int priority = flags & SPF_PRIORITY_MASK;
	int	dest = netbuffer.player;
	
	if(dest >= 0 && dest < MAXPLAYERS)
	{
		dest = clients[dest].jtNetNode;
		if(dest == DCN_MISSING_NODE)
		{
			ST_Message("SendPacket: destnode missing!\n");
			return;
		}
	}
	if(!priority)
	{
		// Non-reliable messages go first by default.
		if(!(flags & SPF_RELIABLE)) priority = 10;
	}
	if(flags & SPF_FRAME)
	{
		// Remove the existing frame packet from the send queue.
		if(lastFrameId != LFID_NONE) 
		{
			//int res = 
			jtNetCancel(lastFrameId);
			//ST_Message( "removing last frame: res=%i\n", res);
		}
	}
	// Send the packet.
	jtNetSend(dest, &netbuffer.msg, netbuffer.headerLength + netbuffer.length,
		(flags & SPF_RELIABLE) != 0, priority, 0, &msgId);
	if(flags & SPF_FRAME)
	{
		// Update the last frame ID.
		lastFrameId = msgId;
	}
}

boolean I_GetPacket()
{
	int from, bytes, i;

	netbuffer.player = -1;
	netbuffer.length = 0;
	bytes = jtNetGet(&from, &netbuffer.msg, sizeof(netbuffer.msg));
	if(!bytes || from < 0) return false;	// No packet.

	// There was a packet!
	netbuffer.length = bytes - netbuffer.headerLength;	
	
	if(server)
	{
		// What is the corresponding player number?
		for(i=1; i<MAXPLAYERS; i++)
			if(clients[i].jtNetNode == from)
			{
				netbuffer.player = i;
				break;
			}
	}
	else 
	{
		// Then it must be from the server.
		netbuffer.player = 0; 
	}
	return true;
}

/*void I_NetCmd()
{
	if(doomcom->command == CMD_SEND)
	{
		localtime++;
        SendPacket(doomcom->remotenode);
	}
	else if(doomcom->command == CMD_GET)
	{
		GetPacket();
    }
}*/

char *I_GetProtocolName(void)
{
	return protocolNames[jtNetGetInteger(JTNET_SERVICE)];
}

// Returns true if the send queue for the player is good for additional
// data. If not, new data shouldn't be sent until the queue has cleared up.
boolean I_CheckSendQueue(int player)
{
	int	num, bytes;

	if(!clients[player].connected || player == consoleplayer) return false;
	num = jtNetCheckQueue(clients[player].jtNetNode, &bytes);
	return (num <= maxQueuePackets);
}


//-----------------------------------------------------------------
//
// Network Events
//
void I_NetServerStart(boolean before)
{
	if(before)
	{
		// The game DLL may have something that needs doing
		// before we actually begin.
		if(gx.NetServerStart) gx.NetServerStart(true);

		// The info strings.
		jtNetSetString(JTNET_SERVER_INFO, serverInfo);
		jtNetSetString(JTNET_NAME, playerName);
	}
	else
	{
		Sv_StartNetGame();

		// The game DLL might want to do something now that the
		// server is started.
		if(gx.NetServerStart) gx.NetServerStart(false);

		if(masterAware)
		{
			I_MAPost(mac_connect);
			I_MAPost(mac_check_connection);
		}
	}
}

void I_NetServerClose(boolean before)
{
	if(before)
	{
		ST_Message("Disconnecting from master.\n");

		// Bye-bye, master server.
		jtNetDisconnectMaster();
		I_MAClear();

		if(gx.NetServerClose) gx.NetServerClose(true);
		Net_StopGame();
	}
	else
	{
		server = false;
	}
	// Call game DLL's NetServerClosed.
	if(gx.NetServerClose) gx.NetServerClose(before);
}

/*void I_NetServerStopping(boolean before)
{
	if(before)
	{
		if(!netgame) return;	// Server can't stop unless we're in netgame.

		// Stop the netgame.
		D_QuitNetGame();
		D_StopNetGame(false);
	}

	// Call game DLL's NetServerStopped.
	if(gx.NetServerStop) gx.NetServerStop(before);
}*/

void I_NetConnecting(boolean before)
{
	if(!before)
	{
		handshake_received = false;
		netgame = true; // Allow sending/receiving of packets.
		server = false;
		I_NetUpdateNodes();
	}

	// Call game DLL's NetConnected.
	gx.NetConnect(before);
}

// before = true if the connection hasn't yet been closed.
void I_NetDisconnect(boolean before)
{
	if(before)
	{
		if(gx.NetDisconnect) gx.NetDisconnect(true);
		Net_StopGame();
	}
	else
	{
		// Call game DLL's NetDisconnected.
		if(gx.NetDisconnect) gx.NetDisconnect(false);
	}
}

// Handle low-level net tick stuff.
void I_NetTicker(void)
{
	masteraction_t act;
	jtnetserver_t info;
	int	i;

	if(!netAvailable) return;

	// Is there a master action to worry about?
	if(I_MAGet(&act))
	{
		switch(act)
		{
		case mac_connect:
			ST_Message( "Connecting to master...\n");
			jtNetSetString(JTNET_MASTER_ADDRESS, masterAddress);
			jtNetSetInteger(JTNET_MASTER_PORT, masterPort);
			// This'll begin the process.
			i = jtNetConnectMaster();
			if(i != JTNET_ERROR_PLEASE_WAIT && i != JTNET_OK)
			{
				ST_Message( "Failure!\n");
				I_MAClear();
				break;
			}
			// Remove this action and move onto the next one.
			I_MARemove();
			break;
			
		case mac_check_connection:
			i = jtNetGetInteger(JTNET_MASTER_CONNECTION);
			if(i == JTNET_OK)
			{
				// Connection has been formed.
				ST_Message( "Connected to master at %s:%i.\n",
					masterAddress, masterPort);
				I_MARemove();
			}
			else if(i != JTNET_ERROR_PLEASE_WAIT)
			{
				ST_Message( "Failed to connect to master at %s:%i.\n", 
					masterAddress, masterPort);
				// Cancel all other actions.
				I_MAClear();
			}
			break;

		case mac_disconnect:
			ST_Message("Disconnecting from master.\n");
			jtNetDisconnectMaster();
			I_MARemove();
			break;

		case mac_get_servers:
			// Send the request for servers.
			ST_Message( "Requesting a list of servers.\n");
			if((i=jtNetMasterGetServer(0, NULL)) != JTNET_OK)
			{
				ST_Message( "Failure! Return code: %i.\n", i);
				I_MAClear();
				break;
			}
			I_MARemove();					
			break;

		case mac_wait_servers:
			// Handle incoming messages.
			jtNetMasterHandler();
			if(jtNetGetInteger(JTNET_EVENT_SERVERLIST_RECEIVED))
			{
				// The list has arrived!
				ST_Message("Server list received.\n");
				I_MARemove();
			}
			break;

		case mac_list_servers:
			for(i=0; ; i++)
			{
				if(jtNetMasterGetServer(i, &info) != JTNET_OK) break;
				if(!i) CON_Printf("    %-20s P/M  L Description:\n", "Name:");
				CON_Printf("%-2d: %-20s %d/%-2d %c %s\n", i, info.name,
					info.players, info.maxPlayers,
					info.canJoin? ' ':'*',
					info.description);
			}
			CON_Printf( "%i server%s found.\n", i, i!=1? "s were" : " was");
			I_MARemove();			
			break;
		}
	}
}

//-----------------------------------------------------------------
// 
// The NET Console Command
//
int CCmdNet(int argc, char **argv)
{
	int		i;
	boolean success = true;

	if(argc == 1)	// No args?
	{
		CON_Printf( "Usage: %s (cmd/args)\n", argv[0]);
		return true;
	}
	if(argc == 2)	// One argument?
	{
		if(!stricmp(argv[1], "shutdown"))
		{
			I_NetShutdown();
			CON_Printf( "Network shut down.\n");
		}
		else if(!stricmp(argv[1], "modems"))
		{
			int num;
			char **modemList = jtNetGetStringList(JTNET_MODEM_LIST, &num);
			if(modemList == NULL)
			{
				CON_Printf( "No modem list available.\n");
			}
			else for(i=0; i<num; i++)
			{
				CON_Printf( "%d: %s\n", i, modemList[i]);
			}
		}
		else if(!stricmp(argv[1], "search"))
		{
			int num = jtNetGetServerInfo(NULL, 0);
			if(num < 0)
			{
				CON_Printf( "Server list is being retrieved. Try again in a moment.\n");
			}
			else if(num == 0)
			{
				CON_Printf( "No server list available.\n");
			}
			else 
			{
				jtnetserver_t *buffer = malloc(sizeof(jtnetserver_t) * num);
				jtNetGetServerInfo(buffer, num);
				CON_Printf( "%d server%s found.\n", num, num!=1? "s were" : " was");
				if(num)
				{
					CON_Printf("%-20s P/M  L Description:\n", "Name:");
					for(i=0; i<num; i++)
						CON_Printf("%-20s %d/%-2d %c %s\n", buffer[i].name,
						buffer[i].players, buffer[i].maxPlayers,
						buffer[i].canJoin? ' ':'*',
						buffer[i].description);
				}
				free(buffer);
			}			
		}
		else if(!stricmp(argv[1], "servers")) 
		{
			if(server)
			{
				CON_Printf( "'net servers' is only for clients.\n");
				return false;
			}
			I_MAPost(mac_connect);
			I_MAPost(mac_check_connection);
			I_MAPost(mac_get_servers);
			I_MAPost(mac_wait_servers);
			I_MAPost(mac_list_servers);
			I_MAPost(mac_disconnect);

/*			// Request the list of servers from the master server.
			if(!jtNetGetInteger(JTNET_EVENT_SERVERLIST_RECEIVED))
			{
				// The list is currently being received!
				CmdReturnValue = false;
				return false;
			}
			if(!I_NetConnectToMaster())	return (CmdReturnValue = false);
			// Send the request for servers.
			CmdReturnValue = (jtNetMasterGetServer(0, NULL) == JTNET_OK);
			if(CmdReturnValue)
				CON_Printf( "Request for server list sent.\n");
			else
				CON_Printf( "Sending of server list request failed.\n");
			return CmdReturnValue;*/
		}
		else if(!stricmp(argv[1], "players"))
		{
			int num = 0;
			char **nameList = jtNetGetStringList(JTNET_PLAYER_NAME_LIST, &num);
			if(!num)
			{
				CON_Printf( "No player list available.\n");
			}
			else 
			{
				CON_Printf( "There %s %d player%s.\n", num==1? "is" : "are", num, num!=1? "s" : "");
				for(i=0; i<num; i++)
					CON_Printf( "%d: %s\n", i, nameList[i]);
			}
		}			
		else if(!stricmp(argv[1], "info"))
		{
			if(server)
			{
				CON_Printf( "Clients:\n");
				for(i=0; i<MAXPLAYERS; i++)
				{
					if(!clients[i].connected) continue;
					CON_Printf( "%i: node %i, node ID %x (ingame:%i)\n", 
						i, clients[i].jtNetNode,
						clients[i].nodeID, players[i].ingame);
				}
			}
			
			CON_Printf( "\nNetwork game: %s\n", netgame? "yes" : "no");
			CON_Printf( "Server: %s\n", server? "yes" : "no");
			//CON_Printf( "Limbo: %s (you're player %d)\n", limbo? "yes" : "no", limboplayer);
			CON_Printf( "TCP/IP address: %s\n", jtNetGetString(JTNET_TCPIP_ADDRESS));
			CON_Printf( "TCP/IP port: %d (0x%x)\n", jtNetGetInteger(JTNET_TCPIP_PORT),
				jtNetGetInteger(JTNET_TCPIP_PORT));
			CON_Printf( "Modem: %d (%s)\n", jtNetGetInteger(JTNET_MODEM),
				jtNetGetString(JTNET_MODEM));
			CON_Printf( "Phone number: %s\n", jtNetGetString(JTNET_PHONE_NUMBER));
			CON_Printf( "Serial: COM %d, baud %d, stop %d, parity %d, flow %d\n",
				jtNetGetInteger(JTNET_COMPORT),
				jtNetGetInteger(JTNET_BAUDRATE),
				jtNetGetInteger(JTNET_STOPBITS),
				jtNetGetInteger(JTNET_PARITY),
				jtNetGetInteger(JTNET_FLOWCONTROL));
		}
		else if(!stricmp(argv[1], "disconnect"))
		{
			if(!netgame)
			{
				CON_Printf( "This client is not connected to a server.\n");
				return false;
			}
			if(server)
			{
				CON_Printf( "You can't disconnect the server! Use \"net server close\".\n");
				return false;
			}

			// Things that have to be done before closing the connection.
			I_NetDisconnect(true);

			success = !jtNetDisconnect();
			if(success)
			{
				CON_Printf( "Disconnected.\n");
				// Things that are done afterwards.
				I_NetDisconnect(false);
			}
			else
				CON_Printf( "Error disconnecting.\n");
		}
		else 
		{
			CON_Printf( "Bad arguments.\n");
			return false; // Bad args.
		}
	}
	if(argc == 3)	// Two arguments?
	{
		if(!stricmp(argv[1], "init"))
		{
			int sp = (!stricmp(argv[2], "tcp/ip") || !stricmp(argv[2], "tcpip"))? JTNET_SERVICE_TCPIP 
				: !stricmp(argv[2], "ipx")? JTNET_SERVICE_IPX 
				: !stricmp(argv[2], "serial")? JTNET_SERVICE_SERIAL
				: !stricmp(argv[2], "modem")? JTNET_SERVICE_MODEM
				: JTNET_SERVICE_UNKNOWN;

			if(sp == JTNET_SERVICE_UNKNOWN)
			{
				ST_Message( "%s is not a supported service provider.\n", argv[2]);
				return false;
			}

			// Initialization.
			success = I_NetInit(sp);
			if(success)
				ST_Message( "Network initialization OK.\n");
			else
				ST_Message( "Network initialization failed!\n");

			// Let everybody know of this.
			CmdReturnValue = success;
		}
		else if(!stricmp(argv[1], "server"))
		{
			/*if(!stricmp(argv[2], "open"))
			{
				if(server) // Is the server already open?
				{
					CON_Printf( "Server already open.\n");
					return false;
				}
				// First set the info strings.
				jtNetSetString(JTNET_SERVER_INFO, serverInfo);
				jtNetSetString(JTNET_NAME, playerName);

				I_NetServerOpening(true);

				// Now we can try to open the server.
				success = !jtNetOpenServer(serverName);
				if(success)
				{
					CON_Printf( "Server \"%s\" opened.\n", serverName);
					I_NetServerOpening(false);
				}
				else
					CON_Printf( "Server opening failed.\n");
				CmdReturnValue = success;
			}
			else */
			if(!stricmp(argv[2], "go") || !stricmp(argv[2], "start"))
			{
				if(netgame)
				{
					CON_Printf( "Already in a netgame.\n");
					return false;
				}
				I_NetServerStart(true);

				// Now we can try to open the server.
				success = !jtNetOpenServer(serverName);
				if(success)
				{
					CON_Printf( "Server \"%s\" started.\n", serverName);
					I_NetServerStart(false);
				}
				else
					CON_Printf( "Server start failed.\n");
				CmdReturnValue = success;
			}
/*			else if(!stricmp(argv[2], "stop"))
			{
				if(!server)
				{
					CON_Printf( "This is not a server!\n");
					return false;
				}
				I_NetServerStopping(true);
				success = !jtNetLockServer(false);
				if(!success)
				{
					CON_Printf( "Couldn't unlock the server.\n");
					return false;
				}
				CON_Printf( "Server stopped, back in limbo mode.\n");
				I_NetServerStopping(false);
			}*/
			else if(!stricmp(argv[2], "close"))
			{
				if(!server)
				{
					CON_Printf( "This is not a server!\n");
					return false;
				}
				// Close the server and kick everybody out.
				I_NetServerClose(true);

				success = !jtNetCloseServer();
				if(success)
				{
					CON_Printf( "Server closed.\n");
					I_NetServerClose(false);
				}
				else
					CON_Printf( "Server closing failed.\n");
			}
			else
			{
				CON_Printf( "Bad arguments.\n");
				return false;
			}
		}
		else if(!stricmp(argv[1], "connect"))
		{	
			char *endptr;
			int idx;

			if(netgame)
			{
				CON_Printf( "This client is already connected.\n");
				return false;
			}

			// Set the name of the player.
			jtNetSetString(JTNET_NAME, playerName);

			I_NetConnecting(true);

			idx = strtol(argv[2], &endptr, 10);
			if(*endptr == 0)	// Conversion ok?
				success = !jtNetConnectIdx(idx);
			else
				success = !jtNetConnect(argv[2]);

			if(success)
			{
				CON_Printf( "Connected to '%s'.\n", argv[2]);
				I_NetConnecting(false);				
			}
			else
				CON_Printf( "Connection failed.\n");

			CmdReturnValue = success;
		}
	}
	if(argc >= 3)
	{
		// TCP/IP settings.
		if(!stricmp(argv[1], "tcpip") || !stricmp(argv[1], "tcp/ip"))
		{
			if(!stricmp(argv[2], "address") && argc > 3)
			{
				success = jtNetSetString(JTNET_TCPIP_ADDRESS, argv[3]);
				CON_Printf( "TCP/IP address: %s\n", argv[3]);
			}
			else if(!stricmp(argv[2], "port") && argc > 3)
			{
				int val = strtol(argv[3], NULL, 0);
				success = jtNetSetInteger(JTNET_TCPIP_PORT, val);
				CON_Printf( "TCP/IP port: %d\n", val);
			}
		}
		// Modem settings.
		if(!stricmp(argv[1], "modem"))
		{
			if(!stricmp(argv[2], "phone"))
			{
				success = jtNetSetString(JTNET_PHONE_NUMBER, argv[3]);
				CON_Printf( "Modem phone number: %s\n", argv[3]);
			}
			else 
			{
				int val = strtol(argv[2], NULL, 0);
				char **modemList = jtNetGetStringList(JTNET_MODEM_LIST, NULL);
				success = jtNetSetInteger(JTNET_MODEM, val);
				if(success)
				{
					CON_Printf( "Selected modem: %d, %s\n", val, modemList[val]);
				}
			}
		}
		// Serial settings.
		else if(!stricmp(argv[1], "serial"))
		{
			if(!stricmp(argv[2], "com"))
			{
				int val = strtol(argv[3], NULL, 0);
				success = jtNetSetInteger(JTNET_COMPORT, val);
				if(success) CON_Printf( "COM port: %d\n", val);
			}
			else if(!stricmp(argv[2], "baud"))
			{
				int val = strtol(argv[3], NULL, 0);
				success = jtNetSetInteger(JTNET_BAUDRATE, val);
				if(success) CON_Printf( "Baud rate: %d\n", val);
			}
			else if(!stricmp(argv[2], "stop"))
			{
				int val = strtol(argv[3], NULL, 0);
				success = jtNetSetInteger(JTNET_STOPBITS, val);
				if(success) CON_Printf( "Stopbits: %d\n", val);
			}
			else if(!stricmp(argv[2], "parity"))
			{
				int val = strtol(argv[3], NULL, 0);
				success = jtNetSetInteger(JTNET_PARITY, val);
				if(success) CON_Printf( "Parity: %d\n", val);
			}
			else if(!stricmp(argv[2], "flow"))
			{
				int val = strtol(argv[3], NULL, 0);
				success = jtNetSetInteger(JTNET_FLOWCONTROL, val);
				if(success) CON_Printf( "Flow control: %d\n", val);
			}
		}
	}
	return success;
}