
//**************************************************************************
//**
//** SV_MISSILE.C
//**
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "dd_def.h"
#include "dd_net.h"
#include "sv_def.h"

// MACROS ------------------------------------------------------------------

#define MAX_MISSILES		64	// This many missiles tracked.
#define MAX_SENT_MISSILES	16	// This many sent to a client at most.
#define MAX_MISSILE_DIST	1024
#define MISSILE_TIMEOUT		35	

// TYPES -------------------------------------------------------------------

typedef struct
{
	boolean used;				
	int type;					
	int flags;					// Combined with type when sent.
	fixed_t	x, y, z;			 
	fixed_t momx, momy, momz;
	sector_t *sector;			

	// This missile must be sent to these players.
	int willSendTo;
	
	// This missile has already been sent to these players.
	int sentTo;					

	// The tic this missile was created (too old missiles will be
	// removed; the player has probably exited or something).
	int time;
} missile_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static missile_t missile_list[MAX_MISSILES];

// CODE --------------------------------------------------------------------

// Find an unused missile entry.
static missile_t *GetMissile()
{
	int	i;
	missile_t *mis;
	
	for(i=0, mis=missile_list; i<MAX_MISSILES; i++, mis++)
	{
		if(mis->used)
		{
			// Has this missile been sent to everybody?
			if(mis->sentTo == mis->willSendTo 
				|| gametic - mis->time > MISSILE_TIMEOUT)
			{
				// This can be replaced.
				mis->used = false;
			}
		}
		if(!mis->used)
		{
			memset(mis, 0, sizeof(*mis));
			mis->used = true;
			mis->time = gametic;
			return mis;
		}
	}
	// Out of missile entries!
	return NULL;
}

void Sv_ScanForMissiles()
{
	thinker_t *th;
	mobj_t *mo;
	missile_t *mis = missile_list;
	int i, sendMask = 0;

	// Compose the sendMask.
	for(i=0; i<MAXPLAYERS; i++) 
		if(clients[i].ready) sendMask |= 1<<i;

	// Check all thinkers.
	for(th=thinkercap.next; th != &thinkercap; th=th->next)
	{
		if(th->function != gx.MobjThinker) continue;
		mo = (mobj_t*) th;
		if(mo->ddflags & (DDMF_DONTDRAW | DDMF_LOCAL)) continue;
		if(mo->ddflags & DDMF_MISSILE && !(mo->ddflags & DDMF_SENT))
		{
			// This will be added.
			mo->ddflags |= DDMF_SENT;
			mis = GetMissile();
			if(!mis) break;	// Too many!
			mis->willSendTo = sendMask;
			mis->type = mo->type;
			mis->flags = 0;
			mis->x = mo->x;
			mis->y = mo->y;
			mis->z = mo->z;
			mis->momx = mo->momx;
			mis->momy = mo->momy;
			mis->momz = mo->momz;
			mis->sector = mo->subsector->sector;
			// XY movement won't be sent for missiles that don't have any.
			if(mis->momx || mis->momy) mis->flags |= DDMS_MOVEMENT_XY;
			if(mis->momz) mis->flags |= DDMS_MOVEMENT_Z;
		}
	}
}
	
void Sv_PackMissiles(int plrNum)
{
	int i, count;
	byte *countPtr = netbuffer.cursor;	
	missile_t *mis;
	mobj_t *plrMo = players[plrNum].mo;
	sector_t *plrSector = players[plrNum].mo->subsector->sector;
	int playerMask = 1 << plrNum;

	// We'll know the number of the missiles when we've checked 
	// through the list.
	Msg_WriteByte(0);

	for(i=0, count=0, mis=missile_list; i<MAX_MISSILES; i++, mis++)
	{
		if(!mis->used || mis->sentTo & playerMask) continue;
		if(!(mis->flags & DDMS_MOVEMENT_XY))
		{
			// This is a 'stationary' missile. We can perform partial 
			// visibility checking.
			if(!P_CheckReject(plrSector, mis->sector)) continue;
			// How about distance?
			if(M_AproxDistance(mis->x - plrMo->x, mis->y - plrMo->y) 
				> MAX_MISSILE_DIST) continue;
		}
		else
		{
			// This missile is moving. We should check where it's heading:
			// missiles too far away but coming closer should be sent.
			
		}
		// Send this missile. Pack it in. 
		// The first word always contains the flags and mobj type.
		Msg_WriteShort(mis->flags | mis->type);
		// Position is always included.
		Msg_WriteShort(mis->x >> 16);
		Msg_WriteShort(mis->y >> 16);
		Msg_WriteShort(mis->z >> 16);
		// How about movement?
		if(mis->flags & DDMS_MOVEMENT_XY)
		{
			Msg_WriteShort(mis->momx >> 8);
			Msg_WriteShort(mis->momy >> 8);
		}
		if(mis->flags & DDMS_MOVEMENT_Z)
		{
			Msg_WriteShort(mis->momz >> 8);
		}
		
		// Mark the missile as sent.
		mis->sentTo |= playerMask;

		// Still room left?
		if(++count == MAX_SENT_MISSILES) break;
	}
	*countPtr = count;
}
