
//**************************************************************************
//**
//** SV_TRACK.C
//**
//** Tracking changes in the world and sending them to clients.
//**
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "dd_def.h"
#include "dd_net.h"
#include "sv_def.h"

// MACROS ------------------------------------------------------------------

#define WRITE_SHORT(byteptr, val)	{(*(short*)(byteptr) = (val)); byteptr += 2;}

// Plane tracker flags.
#define PTF_MOVING		0x01		// Height currently moving.
#define PTF_TEXTURE		0x02		// Texture has been changed.
#define PTF_SOUND_SENT	0x04		// The move sound has been sent.

// Sector tracker flags.
#define STF_LIGHT_CHANGED	0x01	// Light level changed in the sector.

// How long is the largest possible sector update?
#define MAX_SECTORUPD	20
#define MAX_SIDEUPD		9

// TYPES ------------------------------------------------------------------

// Tracker planes.
enum
{
	PT_FLOOR,
	PT_CEILING
};

typedef struct
{
	byte flags;
	fixed_t initial;
	fixed_t dest;
	fixed_t speed;
} plane_tracker_t;

// There's one of these for each sector.
typedef struct
{
	plane_tracker_t	plane[2];	// Floor and ceiling.
	byte flags;
} sector_tracker_t;

typedef struct
{
	unsigned short initialTop, initialMid, initialBottom;
} side_tracker_t;

// EXTERNAL FUNCTIONS -----------------------------------------------------

// PRIVATE DATA -----------------------------------------------------------

static sector_tracker_t	*trackers;
static side_tracker_t *sideTrackers;

// CODE -------------------------------------------------------------------

void Sv_PackSector(int sectornum, int force)
{
	byte buffer[MAX_SECTORUPD], *ptr = buffer;
	sector_t *sec = SECTOR_PTR(sectornum);
	sector_tracker_t *track = &trackers[sectornum];
	plane_tracker_t *pt;
	packet_sectorupdate_t *upd = (packet_sectorupdate_t*) ptr;

	upd->flags = 0;
	upd->sector = sectornum;

	ptr += sizeof(*upd);
	// Check floor first.
	pt = &track->plane[PT_FLOOR];
	if(!force && pt->initial != sec->floorheight || force & DDSU_FLOOR_HEIGHT)
	{
		upd->flags |= DDSU_FLOOR_HEIGHT;
		WRITE_SHORT(ptr, sec->floorheight >> FRACBITS);
	}
	if(!force && pt->flags & PTF_MOVING || force & DDSU_FLOOR_MOVING)
	{
		upd->flags |= DDSU_FLOOR_MOVING;
		// Destination and speed.
		WRITE_SHORT(ptr, pt->dest >> FRACBITS);
		WRITE_SHORT(ptr, (pt->speed&0xffff00) >> 8);
	}
	if(pt->flags & PTF_TEXTURE)
	{
		upd->flags |= DDSU_FLOORPIC;
		WRITE_SHORT(ptr, sec->floorpic);
	}
	// Then ceiling.
	pt = &track->plane[PT_CEILING];
	if(!force && pt->initial != sec->ceilingheight || force & DDSU_CEILING_HEIGHT)
	{
		upd->flags |= DDSU_CEILING_HEIGHT;
		WRITE_SHORT(ptr, sec->ceilingheight >> FRACBITS);
	}
	if(!force && pt->flags & PTF_MOVING || force & DDSU_CEILING_MOVING)
	{
		upd->flags |= DDSU_CEILING_MOVING;
		// Destination and speed.
		WRITE_SHORT(ptr, pt->dest >> FRACBITS);
		WRITE_SHORT(ptr, (pt->speed&0xffff00) >> 8);
	}
	if(pt->flags & PTF_TEXTURE)
	{
		upd->flags |= DDSU_CEILINGPIC;
		WRITE_SHORT(ptr, sec->ceilingpic);
	}
	if(track->flags & STF_LIGHT_CHANGED)
	{	
		upd->flags |= DDSU_LIGHT_LEVEL;
		*ptr++ = sec->lightlevel;
	}
	// If nothing was written, this sector can be skipped.
	if(!upd->flags) return;

	Msg_Write(buffer, ptr-buffer);
}

void Sv_PackSide(int sidenum, int forceflags)
{
	byte buffer[MAX_SIDEUPD], *ptr = buffer;
	side_t *side = SIDE_PTR(sidenum);
	side_tracker_t *st = sideTrackers + sidenum;	
	packet_wallupdate_t *upd = (packet_wallupdate_t*) ptr;

	upd->flags = 0;
	upd->side = sidenum;

	ptr += sizeof(*upd);		
	if((!forceflags && st->initialTop != side->toptexture) 
		|| forceflags & DDWU_TOP)
	{
		upd->flags |= DDWU_TOP;
		WRITE_SHORT(ptr, side->toptexture);
	}
	if((!forceflags && st->initialMid != side->midtexture) 
		|| forceflags & DDWU_MID)
	{
		upd->flags |= DDWU_MID;
		WRITE_SHORT(ptr, side->midtexture);
	}
	if((!forceflags && st->initialBottom != side->bottomtexture) 
		|| forceflags & DDWU_BOTTOM)
	{
		upd->flags |= DDWU_BOTTOM;
		WRITE_SHORT(ptr, side->bottomtexture);
	}
	if(!upd->flags) return;

	Msg_Write(buffer, ptr-buffer);
}

// If speed is zero, the move will be immediate.
// Always give a positive speed (really doesn't matter).
void Sv_PlaneMoves(int sector, boolean ceiling, fixed_t speed)
{
	if(!server) return;

	/*gi.Message( "Sv_PlaneMoves: sec%i, ceil?%i, dest:%f, speed:%f, tex:%i.\n",
		sector, ceiling, FIX2FLT(dest_height), FIX2FLT(speed), texture);*/

	Msg_Begin(psv_sector_update);
	Sv_PackSector(sector, 
		ceiling? (speed? DDSU_CEILING_MOVING : DDSU_CEILING_HEIGHT) 
		: (speed? DDSU_FLOOR_MOVING : DDSU_FLOOR_HEIGHT));
	Net_SendPacket(NSP_BROADCAST, 0);
}

// Texture -1 means 'no change'.
void Sv_TextureChanges(int sidenum, int flags)
{
	if(!server) return;

	/*gi.Message( "Sv_TextureChanges: side%i, top%i, mid%i, bot%i\n",
		sidenum, newTop, newMid, newBottom);*/

	Msg_Begin(psv_wall_update);
	Sv_PackSide(sidenum, flags);
	//gi.SendPacket(DDSP_ALL_PLAYERS, DDPT_WALL_UPDATE, buffer, len);
	Net_SendPacket(NSP_BROADCAST, 0);
}

void Sv_InitTrackers(void)
{
	int	i;
	sector_t *sec;
	side_t *side;

	if(!server) return; // Only if needed.

	trackers = Z_Malloc(sizeof(sector_tracker_t) * numsectors, PU_LEVEL, 0);
	memset(trackers, 0, sizeof(sector_tracker_t) * numsectors);
	// Store the initial heights. No need to send data for nothing.
	for(i=0; i<numsectors; i++)
	{
		sec = SECTOR_PTR(i);
		trackers[i].plane[PT_FLOOR].initial = sec->floorheight;
		trackers[i].plane[PT_CEILING].initial = sec->ceilingheight;
	}

	// Side trackers keep track of changed wall textures.
	sideTrackers = Z_Malloc(sizeof(side_tracker_t) * numsides, PU_LEVEL, 0);
	for(i=0; i<numsides; i++)
	{
		side = SIDE_PTR(i);
		sideTrackers[i].initialMid = side->midtexture;
		sideTrackers[i].initialTop = side->toptexture;
		sideTrackers[i].initialBottom = side->bottomtexture;
	}
}

void Sv_PlaneSound(sector_t *sector, boolean isCeiling, int sound_id, int pace)
{
	int secnum;
	sector_tracker_t *secTrack;
	plane_tracker_t *pTrack;

	if(!netgame || !server) return; // Only if needed.	

	secnum = GET_SECTOR_IDX(sector);
	secTrack = &trackers[secnum];
	pTrack = &secTrack->plane[!isCeiling? PT_FLOOR : PT_CEILING];

	if(!(pTrack->flags & PTF_SOUND_SENT))
	{
		pTrack->flags |= PTF_SOUND_SENT; 
		
		Msg_Begin(psv_plane_sound);
		//WRITE_SHORT(ptr, secnum);
		Msg_WriteShort(secnum);
		Msg_WriteByte((isCeiling? 0x80 : 0) | pace);
		Msg_WriteShort(sound_id);
		//gi.SendPacket(DDSP_ALL_PLAYERS, DDPT_PLANE_SOUND, buffer, ptr-buffer);
		Net_SendPacket(NSP_BROADCAST, 0);
	}
}

// To report a stop in movement, set speed to zero.
void Sv_SectorReport(sector_t *sector, boolean isCeiling, fixed_t dest,
						fixed_t speed, int floorpic)
{
	int secnum;
	sector_tracker_t *secTrack;
	plane_tracker_t *pTrack;
	fixed_t current;

	if(!server) return; // Only if needed.	

	secnum = GET_SECTOR_IDX(sector);
	secTrack = &trackers[secnum];
	pTrack = &secTrack->plane[!isCeiling? PT_FLOOR : PT_CEILING];
	current = isCeiling? sector->ceilingheight : sector->floorheight;

	// Check any changes. 
	if(floorpic != -1)
	{
		// Mark the texture change.
		secTrack->plane[PT_FLOOR].flags |= PTF_TEXTURE;
	}
	if(pTrack->flags & PTF_MOVING)
	{
		// The plane was reported moving the last time.
		if(!speed)
		{
			// The plane has stopped.
			pTrack->flags &= ~(PTF_MOVING | PTF_SOUND_SENT);
			// Send a notification about this.
			Sv_PlaneMoves(secnum, isCeiling, 0);
		}
		else
		{
			if(pTrack->speed != speed || pTrack->dest != dest)
			{
				pTrack->speed = speed;
				pTrack->dest = dest;
				pTrack->flags &= ~PTF_SOUND_SENT;
				// Send a notification about this.
				Sv_PlaneMoves(secnum, isCeiling, speed);
			}
		}
	}
	else
	{
		// The plane was not moving the last time.
		pTrack->speed = speed;
		pTrack->dest = dest;
		if(speed)
		{
			pTrack->flags &= ~PTF_SOUND_SENT;
			pTrack->flags |= PTF_MOVING;
		}
		// Send a notification about this.
		Sv_PlaneMoves(secnum, isCeiling, speed);
	}
}

void Sv_LightReport(sector_t *sector)
{
	int secNum = GET_SECTOR_IDX(sector);
	sector_tracker_t *sTrack = trackers + secNum;

	if(!server) return;

	sTrack->flags |= STF_LIGHT_CHANGED;
	// Send a notification about this. The Light Level flag
	// isn't noticed in forcing, but it prevents the sending
	// of the floor/ceiling data.
	Msg_Begin(psv_sector_update);
	Sv_PackSector(secNum, DDSU_LIGHT_LEVEL);
	//gi.SendPacket(DDSP_ALL_PLAYERS, DDPT_SECTOR_UPDATE, buffer, len);
	Net_SendPacket(NSP_BROADCAST, 0);
}

// Initial world updates are sent reliably.
void Sv_SendWorldUpdate(int playernum)
{
	int		i;

	// Update sectors.
	Msg_Begin(psv_sector_update);
	for(i=0; i<numsectors; i++)
	{
		Sv_PackSector(i, 0);
		// Buffer getting full?
		// Always send after the last one.
		if(Msg_MemoryLeft() < MAX_SECTORUPD	|| i==numsectors-1) 
		{
			// Send out a packet.
			/*gi.SendPacket(playernum | DDSP_RELIABLE, DDPT_SECTOR_UPDATE, 
				buffer, ptr-buffer);*/
			Net_SendPacket(playernum, SPF_RELIABLE);
			// Back to the beginning of the buffer.
			Msg_Begin(psv_sector_update);
		}
	}
	// Update wall textures.
	Msg_Begin(psv_wall_update);
	for(i=0; i<numsides; i++)
	{
		Sv_PackSide(i, 0);
		// Buffer getting full?
		// Always send after the last one.
		if(Msg_MemoryLeft() < MAX_SIDEUPD || i==numsides-1) 
		{
			/*gi.SendPacket(playernum | DDSP_RELIABLE, DDPT_WALL_UPDATE, 
				buffer, ptr-buffer);
			ptr = buffer;*/
			Net_SendPacket(playernum, SPF_RELIABLE);
			Msg_Begin(psv_wall_update);
		}
	}
}
